import tkwant
import kwant
from math import sin, pi
import matplotlib.pyplot as plt


def am_master():
    """Return true for the MPI master rank"""
    return tkwant.mpi.get_communicator().rank == 0


def plot_currents(times, dynamic_current, static_current):
    plt.plot(times, dynamic_current, lw=3, label='dynamic')
    plt.plot([times[0], times[-1]], [static_current] * 2, lw=2, label='static')
    plt.legend(loc=4)
    plt.xlabel(r'time $t$')
    plt.ylabel(r'current $I(t)$')
    plt.show()


def circle(pos):
    (x, y) = pos
    return x ** 2 + y ** 2 < 5 ** 2


def make_system():
    gamma = 1
    lat = kwant.lattice.square(norbs=1)
    syst = kwant.Builder()
    syst[lat.shape(circle, (0, 0))] = 4 * gamma
    syst[lat.neighbors()] = -1 * gamma

    def voltage(site, V_static):
        return 2 * gamma + V_static

    lead = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    lead[lat(0, 0)] = voltage
    lead[lat.neighbors()] = -1 * gamma

    leadr = kwant.Builder(kwant.TranslationalSymmetry((0, -1)))
    leadr[lat(0, 0)] = 2 * gamma
    leadr[lat.neighbors()] = -1 * gamma

    syst.attach_lead(lead)
    syst.attach_lead(leadr)
    return lat, syst


def faraday_flux(time, V_dynamic):
    omega = 0.1
    t_upper = pi / omega
    if 0 <= time < t_upper:
        return V_dynamic * (time - sin(omega * time) / omega) / 2
    return V_dynamic * (time - t_upper / 2)


def tkwant_calculation(syst, operator, tmax, chemical_potential, V_static, V_dynamic):

    params = {'V_static': V_static, 'V_dynamic': V_dynamic}

    # occupation -- for each lead
    occup_left = tkwant.manybody.lead_occupation(chemical_potential + V_static)
    occup_lower = tkwant.manybody.lead_occupation(chemical_potential)
    occupations = [occup_left, occup_lower]

    # Initialize the time-dependent manybody state. Here we use a lower
    # accuracy for the initial adaptive refinement (at time=0) to speed up the calculation.
    # For a realistic simulation, set rtol and atol to smaller values.
    state = tkwant.manybody.State(syst, tmax, occupations, params=params, refine=False)
    state.refine_intervals(rtol=0.05, atol=0.05)

    # loop over time, calculating the current as we go
    expectation_values = []
    times = range(tmax)
    for time in times:
        state.evolve(time)
        # For a realistic simulation, one must set rtol and atol to smaller values,
        # or remove both arguments completely, to converge the manybody integral.
        state.refine_intervals(rtol=0.05, atol=0.05)
        result = state.evaluate(operator)
        expectation_values.append(result)
    return times, expectation_values


def main():

    lat, syst = make_system()

    # add the "dynamic" part of the voltage
    added_sites = tkwant.leads.add_voltage(syst, 0, faraday_flux)
    left_lead_interface = [(a, b)
                           for b in added_sites
                           for a in syst.neighbors(b) if a not in added_sites]
    if am_master():
        kwant.plot(syst, site_color='k', lead_color='grey',
                   hop_lw=lambda a, b: 0.3 if (a, b) in left_lead_interface else 0.1,
                   hop_color=lambda a, b: 'red' if (a, b) in left_lead_interface else 'k')

    syst = syst.finalized()

    # observables
    current_operator = kwant.operator.Current(syst, where=left_lead_interface)

    chemical_potential = 1
    tmax = 100

    # all the voltage applied via the time-dependenece
    V_static = 0
    V_dynamic = 0.5
    times, dynamic_current = tkwant_calculation(syst, current_operator, tmax,
                                                chemical_potential,
                                                V_static, V_dynamic)

    # all the voltage applied statically
    # even though we are doing a "tkwant calculation", we only calculate
    # the results at t=0.
    V_static = 0.5
    V_dynamic = 0
    _, (static_current, *_) = tkwant_calculation(syst, current_operator, 1,
                                                 chemical_potential,
                                                 V_static, V_dynamic)

    if am_master():
        plot_currents(times, dynamic_current, static_current)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2016-2020 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.

from __future__ import print_function  # so we can tell python2 users to upgrade
import os
import sys
import re
import importlib
import collections
import subprocess
import configparser

v = sys.version_info
if v[:2] < (3, 4):
    error = "tkwant requires Python 3.4 or above."
    print(error, file=sys.stderr)
    sys.exit(1)

from setuptools import setup, find_packages, Extension

# Until there is an alternative way to add custom build steps, make sure that
# setuptools' local distutils copy is used.  This works around the removal of
# distutils from stdlib.  See https://github.com/pypa/setuptools/issues/2928.
os.environ['SETUPTOOLS_USE_DISTUTILS'] = 'local'
from distutils.command.build import build
from distutils.errors import DistutilsError, DistutilsModuleError, \
    CCompilerError

from setuptools.command.sdist import sdist
from setuptools.command.build_ext import build_ext

import scipy.integrate

CONFIG_FILE = 'build.conf'
README_FILE = 'README.rst'
README_END_BEFORE = 'See also in this directory:'
REQUIRED_CYTHON_VERSION = (0, 22)
CYTHON_OPTION = '--no-cython'
CYTHON_TRACE_OPTION = '--cython-trace'
MANIFEST_IN_FILE = 'MANIFEST.in'
distr_root = os.path.dirname(os.path.abspath(__file__))

# Let tkwant itself determine its own version.
# We cannot simply import tkwant, as it is not built yet.
spec = importlib.util.spec_from_file_location('_common', 'tkwant/_common.py')
common_module = importlib.util.module_from_spec(spec)
spec.loader.exec_module(common_module)

version = common_module.TKWANT_VERSION


# Cython setup
try:
    sys.argv.remove(CYTHON_OPTION)
    use_cython = False
except ValueError:
    use_cython = True

try:
    sys.argv.remove(CYTHON_TRACE_OPTION)
    trace_cython = True
except ValueError:
    trace_cython = False

if use_cython:
    try:
        import Cython
        from Cython.Build import cythonize
    except ImportError:
        cython_version = ()
    else:
        match = re.match('([0-9.]*)(.*)', Cython.__version__)
        cython_version = [int(n) for n in match.group(1).split('.')]
        # Decrease version if the version string contains a suffix.
        if match.group(2):
            while cython_version[-1] == 0:
                cython_version.pop()
            cython_version[-1] -= 1
        cython_version = tuple(cython_version)



################ utility functions

def complain_cython_unavailable():
    assert not use_cython or cython_version < REQUIRED_CYTHON_VERSION
    if use_cython:
        msg = ("Install Cython {0} or newer so it can be made\n"
               "or use a source distribution of tkwant.")
        ver = '.'.join(str(e) for e in REQUIRED_CYTHON_VERSION)
        print(msg.format(ver), file=sys.stderr)


def extension_config():
    #### Configure external dependencies
    global config_file_present
    config = configparser.ConfigParser()
    try:
        with open(CONFIG_FILE) as f:
            config.read_file(f)
    except IOError:
        config_file_present = False
    else:
        config_file_present = True

    kwrds_by_section = {}
    for section in config.sections():
        kwrds_by_section[section] = kwrds = {}
        for name, value in config.items(section):
            kwrds[name] = value.split()
    return kwrds_by_section


def _successful_link(libs):
    cmd = ['gcc']
    cmd.extend(['-l' + lib for lib in libs])
    cmd.extend(['-o/dev/null', '-xc', '-'])
    try:
        p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError:
        return False
    else:
        p.communicate(input=b'int main() {}\n')
        return p.wait() == 0


def merge_dicts(*dicts):
    result = collections.defaultdict(list)
    for d in dicts:
        for key, item in d.items():
            result[key] = list(set(result[key] + item))
    return result


def extensions():
    """Return a list of tuples (args, kwrds) to be passed to Extension."""
    global build_summary
    build_summary = []
    result = []
    kwrds_by_section = extension_config()

    ### add tkwant components

    # kernels
    result.append((['tkwant.onebody.kernels', ['tkwant/onebody/kernels.pyx']],
                   {}))

    # solvers
    result.append((['tkwant.onebody.solvers',
      ['tkwant/onebody/solvers.pyx']],
     {
      'depends': ['tkwant/onebody/kernels.pxd'],
     }
    ))


    #### Add cython tracing macro
    if trace_cython:
        for args, kwargs in result:
            macros = kwargs.get('define_macros', [])
            macros.append(('CYTHON_TRACE', '1'))
            kwargs['define_macros'] = macros

    build_summary = '\n'.join(build_summary)
    return result


def ext_modules(extensions):
    """Prepare the ext_modules argument for setuptools.

    If Cython is not to be run, replace .pyx extensions with .c or .cpp, and
    check timestamps.
    """
    if use_cython and cython_version >= REQUIRED_CYTHON_VERSION:
        return cythonize([Extension(*args, **kwrds) for args, kwrds in extensions],
                         language_level=3,
                         compiler_directives={'profile': False,
                                              'linetrace': trace_cython}
                        )

    # Cython is not going to be run: replace pyx extension by that of
    # the shipped translated file.

    result = []
    problematic_files = []
    for args, kwrds in extensions:
        name, sources = args

        language = kwrds.get('language')
        if language is None:
            ext = '.c'
        elif language == 'c':
            ext = '.c'
        elif language == 'c++':
            ext = '.cpp'
        else:
            print('Unknown language: {}'.format(language), file=sys.stderr)
            exit(1)

        pyx_files = []
        cythonized_files = []
        new_sources = []
        for f in sources:
            if f.endswith('.pyx'):
                pyx_files.append(f)
                f = f.rstrip('.pyx') + ext
                cythonized_files.append(f)
            new_sources.append(f)
        sources = new_sources

        # Complain if cythonized files are older than Cython source files.
        try:
            cythonized_oldest = min(os.stat(f).st_mtime
                                    for f in cythonized_files)
        except OSError:
            print("error: Cython-generated file {} is missing.".format(f),
                  file=sys.stderr)
            complain_cython_unavailable()
            exit(1)

        for f in pyx_files + kwrds.get('depends', []):
            if f == CONFIG_FILE:
                # The config file is only a dependency for the compilation
                # of the cythonized file, not for the cythonization.
                continue
            if os.stat(f).st_mtime > cythonized_oldest:
                problematic_files.append(f)

        result.append(Extension(name, sources, **kwrds))

    if problematic_files:
        problematic_files = ", ".join(problematic_files)
        msg = ("Some Cython source files are newer than files that should have\n"
               "been derived from them, but {}.\n"
               "\n"
               "Affected files: {}")
        if use_cython:
            if not cython_version:
                reason = "Cython is not installed"
            else:
                reason = "the installed Cython is too old"
            print(banner(" Error "), msg.format(reason, problematic_files),
                  banner(), sep="\n", file=sys.stderr)
            print()
            complain_cython_unavailable()
            exit(1)
        else:
            reason = "the option {} has been given or no cython available.".format(CYTHON_OPTION)
            dontworry = ('(Do not worry about this if you are building tkwant\n'
                         'from unmodified sources, e.g. with "pip install".)\n')
            print(banner(" Caution "), dontworry,
                  msg.format(reason, problematic_files),
                  banner(), sep='\n', file=sys.stderr)

    return result


def write_version(fname):
    # This could be a hard link, so try to delete it first.  Is there any way
    # to do this atomically together with opening?
    try:
        os.remove(fname)
    except OSError:
        pass
    with open(fname, 'w') as f:
        f.write("# This file has been created by setup.py.\n")
        f.write("version = '{}'\n".format(version))


def long_description():
    text = []
    try:
        with open(README_FILE) as f:
            for line in f:
                if line.startswith(README_END_BEFORE):
                    break
                text.append(line.rstrip())
            while text[-1] == "":
                text.pop()
    except:
        return ''
    return '\n'.join(text)


def git_lsfiles():
    try:
        p = subprocess.Popen(['git', 'ls-files'], cwd=distr_root,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError:
        return

    if p.wait() != 0:
        return
    return p.communicate()[0].decode().split('\n')[:-1]


def banner(title=''):
    starred = title.center(79, '*')
    return '\n' + starred if title else starred



################ build commands

class tkwant_build(build):

    def run(self):
        super().run()


error_msg = """{header}
The compilation of tkwant has failed.  Please examine the error message
above and consult the installation instructions in README.rst.
You might have to customize {{file}}.

Build configuration was:

{{summary}}
{sep}
"""
error_msg = error_msg.format(header=banner(' Error '), sep=banner())

class tkwant_build_ext(build_ext):
    def run(self):
        if not config_file_present:
            # Create an empty config file if none is present so that the
            # extensions will not be rebuilt each time.  Only depending on the
            # config file if it is present would make it impossible to detect a
            # necessary rebuild due to a deleted config file.
            with open(CONFIG_FILE, 'w') as f:
                f.write('# Created by setup.py - feel free to modify.\n')

        try:
            build_ext.run(self)
        except (DistutilsError, CCompilerError):
            print(error_msg.format(file=CONFIG_FILE, summary=build_summary),
                  file=sys.stderr)
            raise
        print(banner(' Build summary '))
        print(build_summary)
        print(banner())


class tkwant_sdist(sdist):
    sub_commands = [('build', None)] + sdist.sub_commands

    def run(self):
        """
        Create MANIFEST.in from git if possible, otherwise check that MANIFEST.in
        is present.

        Right now (2015) generating MANIFEST.in seems to be the only way to
        include files in the source distribution that setuptools does not think
        should be there.  Setting include_package_data to True makes setuptools
        include *.pyx and other source files in the binary distribution.
        """
        manifest = os.path.join(distr_root, MANIFEST_IN_FILE)
        names = git_lsfiles()
        if names is None:
            if not (os.path.isfile(manifest) and os.access(manifest, os.R_OK)):
                print("Error:", MANIFEST_IN_FILE,
                      "file is missing and Git is not available"
                      " to regenerate it.", file=sys.stderr)
                exit(1)
        else:
            with open(manifest, 'w') as f:
                for name in names:
                    a, sep, b = name.rpartition('/')
                    if b == '.gitignore':
                        continue
                    stem, dot, extension = b.rpartition('.')
                    f.write('include {}'.format(name))
                    if extension == 'pyx':
                        f.write(''.join([' ', a, sep, stem, dot, 'c']))
                    f.write('\n')

        sdist.run(self)

        if names is None:
            msg = ("Git was not available to generate the list of files "
                   "to be included in the\n"
                   "source distribution.  The old {} was used.")
            msg = msg.format(MANIFEST_IN_FILE),
            print(banner(' Caution '), msg, banner(), sep='\n', file=sys.stderr)

    def make_release_tree(self, base_dir, files):
        sdist.make_release_tree(self, base_dir, files)



################ requirements

requirements = (
    "cython>=0.21.1",
    "numpy>=1.8.2",
    "scipy>=0.14.0",
    "sympy",
    "kwant>=1.3,<2.0",
    "mpi4py>=2.0",
    "tinyarray",
    "kwantspectrum",
)

test_requirements = (
    "pytest",
)

setup_requirements = (
    "pytest-runner",
)

classifiers = """\
    Development Status :: 5 - Production/Stable
    Intended Audience :: Science/Research
    Programming Language :: Python :: 3 :: Only
    Topic :: Scientific/Engineering
    Topic :: Scientific/Engineering :: Physics
    Operating System :: POSIX
    Operating System :: Unix
    Operating System :: MacOS :: MacOS X
    Operating System :: Microsoft :: Windows"""


def main():
    setup(name='tkwant',
          version=version,
          author='tkwant authors',
          author_email='tkwant-authors@kwant-project.org',
          description='Package for time-dependent quantum transport simulations',
          long_description=long_description(),
          platforms=['Unix', 'Linux', 'Mac OS-X', 'Windows'],
          url="https://tkwant.kwant-project.org/",
          license='BSD',
          packages=find_packages('.'),
          cmdclass={'build': tkwant_build,
                    'sdist': tkwant_sdist,
                    'build_ext': tkwant_build_ext},
          ext_modules=ext_modules(extensions()),
          classifiers=[c.strip() for c in classifiers.split('\n')],
          install_requires=requirements,
          tests_require=test_requirements,
          setup_requires=setup_requirements
         )


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright 2022 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.

"""Test module for `tkwant.greenfunctions` and `tkwant.special`"""

from .. import special, manybody, system
import kwant

import pytest
import numpy as np
from numpy.testing import (assert_array_almost_equal, assert_almost_equal)


def make_double_impurity_matrices(gamma_L, gamma_C, gamma_R, epsilon_L, epsilon_R, mu_L, mu_R):
    h_ss = np.array([[epsilon_L, -gamma_C], [-gamma_C, epsilon_R]])
    h_se = np.array([[- gamma_L, 0], [0, - gamma_R]])
    gr_e = np.array([special.g_ee_retarded(mu_L), special.g_ee_retarded(mu_R)])
    mu_e = [mu_L, mu_R]
    return h_ss, h_se, gr_e, mu_e


def make_double_impurity_system(gamma_L, gamma_C, gamma_R, eps_L, eps_R):

    # system building
    lat = kwant.lattice.chain(a=1, norbs=1)
    syst = kwant.Builder()

    # central scattering region
    syst[(lat(x) for x in [-1, 2])] = 0
    syst[lat(0)] = eps_L
    syst[lat(1)] = eps_R

    syst[(lat(0), lat(-1))] = -gamma_L
    syst[(lat(1), lat(0))] = -gamma_C
    syst[(lat(2), lat(1))] = -gamma_R

    # add leads
    sym = kwant.TranslationalSymmetry((-1,))
    lead_left = kwant.Builder(sym)
    lead_left[lat(0)] = 0
    lead_left[lat.neighbors()] = -1
    syst.attach_lead(lead_left)
    syst.attach_lead(lead_left.reversed())

    return syst, lat


def create_system(length):

    def onsite_potential(site, time):
        """Time dependent onsite potential (static part + V(t))"""
        return 1 + v(time)

    # system building
    lat = kwant.lattice.square(a=1, norbs=1)
    syst = kwant.Builder()

    # central scattering region
    syst[(lat(x, 0) for x in range(length))] = 0
    syst[lat.neighbors()] = -1

    # add leads
    sym = kwant.TranslationalSymmetry((-1, 0))
    lead_left = kwant.Builder(sym)
    lead_left[lat(0, 0)] = 0
    lead_left[lat.neighbors()] = -1
    syst.attach_lead(lead_left)
    syst.attach_lead(lead_left.reversed())

    return syst


def test_green_quantumdot_flatband():

    # couplings
    gamma_L = 0.14
    gamma_C = 0.05
    gamma_R = 0.08

    # onsite elements
    epsilon_L = 0.5
    epsilon_R = 0.3

    # chemical potentials
    mu_L = 0
    mu_R = 0

    h_ss, h_se, gr_e, mu_e = make_double_impurity_matrices(gamma_L, gamma_C, gamma_R, epsilon_L, epsilon_R, mu_L, mu_R)

    hybRe = np.array([[0, 0], [0, 0]])
    hybIm = np.array([[- gamma_L * gamma_L, 0], [0, - gamma_R * gamma_R]])

    green_flatband = special.GreenFlatBand(h_ss, h_se, gr_e, mu_e)
    green_flatband_eq = special.GreenFlatBandEq.from_block_hamiltonian(h_ss, h_se, gr_e)
    green_flatband_alt = special.GreenFlatBandEq(h_ss, hybRe, hybIm)

    times = np.linspace(-20, 20, 21)

    for i in [0, 1]:
        for j in [0, 1]:
            for time in times:

                gless = green_flatband.lesser(i, j, time)
                gless_eq = green_flatband_eq.lesser(i, j, time)
                gless_alt = green_flatband_alt.lesser(i, j, time)

                assert_almost_equal(gless, gless_eq)
                assert_almost_equal(gless, gless_alt)

                ggreat = green_flatband.greater(i, j, time)
                ggreat_eq = green_flatband_eq.greater(i, j, time)
                ggreat_alt = green_flatband_alt.greater(i, j, time)

                assert_almost_equal(ggreat, ggreat_eq)
                assert_almost_equal(ggreat, ggreat_alt)


@pytest.mark.integtest
def test_green_retarded():

    # when not all leads are taken into account, this should also hold for the retarded function

    # parameters
    tmax = 20

    # create system
    syst = create_system(5).finalized()

    green = manybody.GreenFunction(syst, tmax, occupations=manybody.lead_occupation(chemical_potential=0))
    green_gen = manybody.greenfunctions._GreenGeneric(syst, tmax, occupations=manybody.lead_occupation(chemical_potential=None))
    assert_almost_equal(green.retarded(0, 0), -green_gen.evaluate(0, 0))

    green = manybody.GreenFunction(syst, tmax, occupations=[manybody.lead_occupation(chemical_potential=0), None])
    green_gen = manybody.greenfunctions._GreenGeneric(syst, tmax, occupations=[manybody.lead_occupation(chemical_potential=None), None])
    assert_almost_equal(green.retarded(0, 0), -green_gen.evaluate(0, 0))


@pytest.mark.integtest
def test_equal_time_green_with_density_current():

    # couplings
    gamma_L = 0.14
    gamma_C = 0.05
    gamma_R = 0.08

    # onsite elements
    epsilon_R = 0.3

    # chemical potentials
    mu_L = 0.5
    mu_R = - 0.8

    n0_wave = []
    n0_green = []

    i10_wave = []
    i10_green = []

    epsilons = np.linspace(-1, 1, 3)

    for epsilon_L_ in epsilons:

        syst, lat = make_double_impurity_system(gamma_L, gamma_C, gamma_R, epsilon_L_, epsilon_R)
        syst = syst.finalized()

        occupations = [manybody.lead_occupation(chemical_potential=mu_L),
                       manybody.lead_occupation(chemical_potential=mu_R)]

        # -- via wavefunction and density operator

        density_operator = kwant.operator.Density(syst, where=[lat(0)])
        current_operator = kwant.operator.Current(syst, where=[(lat(1), lat(0))])

        state = manybody.State(syst, tmax=1, occupations=occupations)
        density = state.evaluate(density_operator)
        current = state.evaluate(current_operator)
        n0_wave.append(density)
        i10_wave.append(current)

        # -- via greenfunction from tkwant

        idx = system.siteId(syst, lat)
        site_0 = idx(0)
        site_1 = idx(1)

        green = manybody.GreenFunction(syst, tmax=1, occupations=occupations)
        g_less_00 = green.lesser(site_0, site_0)
        g_less_10 = green.lesser(site_1, site_0)

        density = g_less_00.imag
        current = - 2 * gamma_C * g_less_10.real
        n0_green.append(density)
        i10_green.append(current)

        assert_array_almost_equal(n0_wave, n0_green)
        assert_array_almost_equal(i10_wave, i10_green)


@pytest.mark.integtest
def test_green_numeric_vs_flatband():

    green_types = {}
    green_types['lesser'] = 'G^<'
    green_types['greater'] = 'G^>'
    green_types['ordered'] = 'G^T'
    green_types['anti_ordered'] = 'G^{Tbar}'
    green_types['retarded'] = 'G^R'
    green_types['advanced'] = 'G^A'
    green_types['keldysh'] = 'G^K'

    green_data_00 = {key: [] for key in green_types.keys()}
    green_data_01 = {key: [] for key in green_types.keys()}
    green_data_10 = {key: [] for key in green_types.keys()}
    green_data_11 = {key: [] for key in green_types.keys()}

    # couplings
    gamma_L = 0.14
    gamma_C = 0.05
    gamma_R = 0.08

    # onsite elements
    epsilon_L = 0.5
    epsilon_R = 0.3

    # chemical potentials
    mu_L = 0.5
    mu_R = - 0.8

    syst, lat = make_double_impurity_system(gamma_L, gamma_C, gamma_R, epsilon_L, epsilon_R)
    syst = syst.finalized()

    idx = system.siteId(syst, lat)
    site_0 = idx(0)
    site_1 = idx(1)

    occupations = [manybody.lead_occupation(chemical_potential=mu_L),
                   manybody.lead_occupation(chemical_potential=mu_R)]

    times = np.linspace(0, 20, 5)

    # -- negative times
    green = manybody.GreenFunction(syst, max(times), occupations)

    for time in times[1:]:
        green.evolve(0, time)
        green.refine_intervals()

        for gtype in green_types.keys():
            green_data_00[gtype].insert(0, getattr(green, gtype)(site_0, site_0))
            green_data_01[gtype].insert(0, getattr(green, gtype)(site_0, site_1))
            green_data_10[gtype].insert(0, getattr(green, gtype)(site_1, site_0))
            green_data_11[gtype].insert(0, getattr(green, gtype)(site_1, site_1))

    # -- positive times
    green = manybody.GreenFunction(syst, max(times), occupations)

    for time in times:
        green.evolve(time, 0)
        green.refine_intervals()

        for gtype in green_types.keys():
            green_data_00[gtype].append(getattr(green, gtype)(site_0, site_0))
            green_data_01[gtype].append(getattr(green, gtype)(site_0, site_1))
            green_data_10[gtype].append(getattr(green, gtype)(site_1, site_0))
            green_data_11[gtype].append(getattr(green, gtype)(site_1, site_1))

    times_pm = np.concatenate((-times[:0:-1], times))
    green_data_00 = {key: np.array(val) for key, val in green_data_00.items()}
    green_data_01 = {key: np.array(val) for key, val in green_data_01.items()}
    green_data_10 = {key: np.array(val) for key, val in green_data_10.items()}
    green_data_11 = {key: np.array(val) for key, val in green_data_11.items()}

    matr = make_double_impurity_matrices(gamma_L, gamma_C, gamma_R, epsilon_L, epsilon_R, mu_L, mu_R)

    green_flatband = special.GreenFlatBand(*matr)

    green_data_00_ref = {key: np.array([getattr(green_flatband, key)(0, 0, t) for t in times_pm])
                         for key in green_types.keys()}
    green_data_01_ref = {key: np.array([getattr(green_flatband, key)(0, 1, t) for t in times_pm])
                         for key in green_types.keys()}
    green_data_10_ref = {key: np.array([getattr(green_flatband, key)(1, 0, t) for t in times_pm])
                         for key in green_types.keys()}
    green_data_11_ref = {key: np.array([getattr(green_flatband, key)(1, 1, t) for t in times_pm])
                         for key in green_types.keys()}

    for i, (key, gtype) in enumerate(green_types.items()):

        assert_array_almost_equal(green_data_00[key], green_data_00_ref[key], decimal=1)
        assert_array_almost_equal(green_data_01[key], green_data_01_ref[key], decimal=1)
        assert_array_almost_equal(green_data_10[key], green_data_10_ref[key], decimal=1)
        assert_array_almost_equal(green_data_11[key], green_data_11_ref[key], decimal=1)



# -*- coding: utf-8 -*-
#
# cython: embedsignature=True
#
# Copyright 2016, 2017 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.


cdef class Kernel:
    cdef public int size, nevals



# Copyright 2016 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.

"""Simple sphinx extension that allows for a section that can be
hidden/shown on click using javascript"""

from docutils import nodes
from docutils.parsers.rst import directives, Directive

id_count = 0

class specialnote(nodes.General, nodes.Element):
    pass

class SpecialNote(Directive):

    required_arguments = 1
    final_argument_whitespace = True
    has_content = True

    def run(self):
        global id_count

        self.assert_has_content()
        text = '\n'.join(self.content)

        classes = directives.class_option(self.arguments[0])

        node = specialnote(text)

        node['classes'].extend(classes)
        node['title'] = self.arguments[0]
        node['myid'] = "specialnote-id" + str(id_count)
        id_count = id_count + 1

        self.state.nested_parse(self.content, self.content_offset, node)

        return [node]

def visit_html(self, node):
    self.body.append("<div class=\"specialnote-title\" id=\"" + node['myid'] +
                     "-title\">" + node['title'])
    self.body.append("</div>")
    self.body.append("<div class=\"specialnote-body\" id=\"" +
                     node['myid'] + "-body\">")

def leave_html(self, node):
    self.body.append("</div>\n")
    #If javascript is enabled, hide the content by default. Otherwise show.
    self.body.append("<script language='javascript'>")
    self.body.append("$('" + "#" + node['myid'] +
                     "-title').append(\" <a style='float: right;' id=\\\"" +
                     node['myid'] +
                     "-button\\\" href=\\\"javascript:togglediv('" +
                     node['myid'] + "')\\\">show</a>\");")
    self.body.append("$('" + "#" + node['myid'] +
                     "-body').css('display', 'none');")
    self.body.append("</script>\n")

def visit_latex(self, node):
    self.body.append("\strong{" + node['title'] +"}\n\n")

def leave_latex(self, node):
    pass

def setup(app):
    app.add_node(specialnote,
                 html=(visit_html, leave_html),
                 latex=(visit_latex, leave_latex))

    app.add_directive('specialnote', SpecialNote)



from cmath import exp
from math import cos, sqrt, pi
import numpy as np
import matplotlib.pyplot as plt

import kwant
import tkwant


def make_system(a=1, gamma=1.0, radius=10, radius_time_dep=3):
    """Make a tight binding system on a single square lattice"""
    # `a` is the lattice constant and `gamma` the hopping integral
    # both set by default to 1 for simplicity.

    lat = kwant.lattice.square(a, norbs=1)
    syst = kwant.Builder()

    def circle(r):
        def inner(pos):
            (x, y) = pos
            return x ** 2 + y ** 2 < r ** 2
        return inner

    def hopx(site1, site2, magnetic_field_strength):
        y = site1.pos[1]
        return -gamma * exp(-1j * magnetic_field_strength * y)

    def potential(site, time, voltage):
        x, y = site.pos
        return 4 * gamma + sqrt(x**2 + y**2) * voltage * (1 - cos(time))

    def side_color_func(site):
            time_dep = circle(radius_time_dep)
            return 'r' if time_dep(site.pos) else 'k'

    # Define the quantum dot
    syst[lat.shape(circle(radius), (0, 0))] = 4 * gamma
    syst[lat.shape(circle(radius_time_dep), (0, 0))] = potential
    # hoppings in x-direction
    syst[kwant.builder.HoppingKind((1, 0), lat, lat)] = hopx
    # hoppings in y-directions
    syst[kwant.builder.HoppingKind((0, 1), lat, lat)] = -gamma

    # Plot system
    kwant.plot(syst, site_color=side_color_func)

    # It's a closed system, so no leads
    return syst


def main():

    # construct a tight binding system and plot it
    syst = make_system().finalized()

    # set parameters
    params = {'voltage': 1.0, 'magnetic_field_strength': 1.0}

    tparams = params.copy()
    tparams['time'] = 0  # the initial time

    # create an observable for calculating the average radius of a wavefunction
    def radius(site):
        x, y = site.pos
        return sqrt(x**2 + y**2)

    density_operator = kwant.operator.Density(syst, radius, sum=True)

    # create a time-dependent wavefunction that starts in the ground state
    hamiltonian = syst.hamiltonian_submatrix(params=tparams)
    eigenvalues, eigenvectors = np.linalg.eigh(hamiltonian)

    ground_state = tkwant.onebody.WaveFunction.from_kwant(syst=syst,
                                                          psi_init=eigenvectors[:, 0],
                                                          energy=eigenvalues[0],
                                                          params=params)

    # evolve forward in time, calculating the average radius of the wavefunction
    times = np.arange(0, 10 * pi, 0.1)
    average_radius = []
    for time in times:
        ground_state.evolve(time)
        density = ground_state.evaluate(density_operator)
        average_radius.append(density)

    # plot the result
    plt.plot(times, average_radius)
    plt.xlabel(r'time $t$')
    plt.ylabel(r'average radius $n(t)$')
    plt.show()


if __name__ == '__main__':
    main()



# Script 2/2 to produce the figure in Tkwant's "DC current through a 1D chain" tutorial.
#
# Step 1: run the simulation (chem_vs_elec_bias_run_computation.py)
# Step 2: plot the data (chem_vs_elec_bias_plot_results.py)
#
# This is the script for step 2 to plot the result.
# Make sure that the file containing the result data
# (chem_vs_elec_bias_result_data.npy, produced by running the script for step 1)
# is placed in the same directory as this script.

import matplotlib.pyplot as plt
import pickle
import pylab
import numpy as np

# --- plot settings
markersize = 14

fac = 6
pylab.rcParams['figure.figsize'] = (fac * np.sqrt(2), fac)

plt.rcParams.update({'font.size': 15})
plt.rc('text', usetex=True)
plt.rc('font', family='serif')

# --- load data
data = pickle.load(open('./chem_vs_elec_bias_result_data.npy', 'rb'))

# --- plot and store figure
fig, (ax1, ax2) = plt.subplots(2)

for v, res in data.items():
    for ax in [ax1, ax2]:
        ax.plot(v, res['I_vc_sc'], 'x', color='#4e79a7')
        ax.plot(v, res['I_vc_wf'], '+', color='#4e79a7')
        ax.plot(v, res['I_ve_sc'], 'x', color='#f28e2b')
        ax.plot(v, res['I_ve_wf'], '+', color='#f28e2b')
        ax.plot(v, res['I_ve_wf_tr'], '.', color='#f28e2b')

for ax in [ax1, ax2]:
    ax.plot(v, res['I_vc_sc'], 'x', color='#4e79a7', label='(Ia) chem drop, scat')
    ax.plot(v, res['I_vc_wf'], '+', color='#4e79a7', label='(Ib) chem drop, wf')
    ax.plot(v, res['I_ve_sc'], 'x', color='#f28e2b', label='(IIa) el drop, scat')
    ax.plot(v, res['I_ve_wf'], '+', color='#f28e2b', label='(IIb) el drop, wf')
    ax.plot(v, res['I_ve_wf_tr'], '.', color='#f28e2b', label='(IIc) el drop, wf gauge')

    ax.set_ylabel(r'$I (e \gamma / h)$')
    ax.legend(loc=5)

ax2.set_xlabel(r'$V (\gamma)$')

ax1.set_xscale('log')
ax1.set_yscale('log')
ax1.set_xlim([5e-7, 1.5e-1])
ax1.set_ylim([5e-7, 1.5e-1])

plt.savefig("chem_vs_elec_bias_current.png", bbox_inches='tight')

plt.show()



# Copyright 2016-2022 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.
"""Tools for solving the one-body time-dependent Schrödinger equation."""

import collections.abc
import functools
from cmath import exp
import numpy as np
import scipy.sparse as sp

import kwant
import kwantspectrum

from .. import leads, _common, _logging, system
from . import kernels, solvers

__all__ = ['WaveFunction', 'ScatteringStates', 'Task', 'make_extended_system',
           '_get_time_name_from_wavefunction']


# set module logger
logger = _logging.make_logger(name=__name__)
log_func = _logging.log_func(logger)


# data formats

class Task:
    """Data format to store the set of quantum numbers that uniquely indentifies
    a onebody state and the weight of that state in the manybody sum.

    Attributes
    ----------
    lead : int
        lead index
    mode : int
        scattering mode index
    energy : float
        energy of the onebody state
    momentum : float
        momentum of the onebody state
    weight : float or numpy float array
        weighting factor of the one-body state in the manybody sum
        weigth = math_weight * phys_weight, where math_weight is the weighting
        factor from the integration quadrature
    math_weight : float or numpy float array
        mathematical weighting factor from the numerical quadrature rule
    phys_weight : float
        physical weighting factor (fermi function, pi factors..)
    """
    def __init__(self, lead: int, mode: int, energy: float, weight: float,
                 math_weight: float = None, phys_weight: float = None,
                 momentum: float = None):
        self.lead = lead
        self.mode = mode
        self.energy = energy
        self.momentum = momentum
        self.weight = weight
        self.math_weight = math_weight
        self.phys_weight = phys_weight

    def __eq__(self, other) -> bool:
        if not isinstance(other, Task):
            return NotImplemented
        return (
            (self.lead, self.mode, self.energy, self.momentum,
             self.weight, self.math_weight, self.phys_weight) ==
            (other.lead, other.mode, other.energy, other.momentum,
             other.weight, other.math_weight, other.phys_weight))

    def __str__(self):
        string = "onebody task: lead={lead}, mode={mode}, " \
                 "energy={energy}, momentum={momentum}, " \
                 "weight={weight}, math_weight={math_weight}, " \
                 "phys_weight={phys_weight}".format(**self.__dict__)
        return string


def _operator_bound(operator):
    """Return True iff operator is bound"""
    try:
        return bool(operator._bound_onsite or operator._bound_hamiltonian)
    except AttributeError:
        if hasattr(operator, '_bound_onsite'):
            return bool(operator._bound_onsite)
        if hasattr(operator, '_bound_hamiltonian'):
            return bool(operator._bound_hamiltonian)
        return False


def _get_time_name_from_wavefunction(wavefunction_type=None, derived_type=None):
    """Return name, initial time and onebody wave function type"""
    default_arg = _common.get_default_function_argument
    try:
        if wavefunction_type is None:
            wavefunction_type = default_arg(derived_type, 'wavefunction_type')
        time_name = default_arg(wavefunction_type, 'time_name')
        time_start = default_arg(wavefunction_type, 'time_start')
    except Exception:
        time_name = _common.time_name
        time_start = _common.time_start
        wavefunction_type = None
        logger.warning('retrieving initial time and time argument name from',
                       'the onebody wavefunction failed, use default values: ',
                       '"time_name"={}, "time_start"={}'.format(time_name,
                                                                time_start))
    return time_name, time_start, wavefunction_type

class _PerturbationWrap:
    def __init__(self, wt):
        self.wt = wt
        self.qt = None  # this will hold the interaction
        self._perturb_set = False

    def size(self):
        return wt.size

    def set(self, qt):
        self.qt = qt
        self._perturb_set = True

    def evaluate(self, time):
        wt = self.wt.evaluate(time)
        try:
            qt = self.qt(time)
            return wt + qt
        except Exception:
            return wt

    def apply(self, time, ket, out):
        try:
            self.wt.apply(time, ket, out)
        except Exception as exept:
            if self.wt is None:
                pass
            else:
                raise exept
        if self._perturb_set:
            qt = self.qt(time)
            out[:] += qt * ket
        else:
            return



def make_extended_system(syst, solver_type, boundaries=None, tparams=None):

    tmax = None
    if syst.leads:
        if not boundaries:
            raise ValueError('Must provide boundary conditions for '
                             'systems with leads.')
        H0, boundary_slices = system._hamiltonian_with_boundaries(syst, boundaries,
                                                                  tparams)
        # if any boundary has a tmax, then take it. if the tmax are different,
        # we will always take the lowest for safety
        for bdr in boundaries:
            if bdr.tmax is not None:
                if tmax is not None:
                    tmax = min(tmax, bdr.tmax)
                else:
                    tmax = bdr.tmax
    else:
        if boundaries is not None:
            logger.warning('boundaries will not be used for extended system')
        H0 = syst.hamiltonian_submatrix(params=tparams, sparse=True)
        boundary_slices = None

    size = H0.shape[0]
    assert _common.is_type(size, 'integer')
    solver = solver_type(size)
    work = np.zeros((size,), dtype=complex)

    return system.ExtendedSystem(syst, H0, boundary_slices, solver, work, tmax)


class WaveFunction:
    r"""A class to solve the time-dependent single-particle wavefunction.

    The time-dependent single-particle Schrödinger equation is

        :math:`i \partial_t \psi = (H_0 + W(t)) \psi`,

    where the total Hamiltonian :math:`H(t) = H_0 + W(t)` has been splitted
    into a static part :math:`H_0` and a time-dependent perturbation
    :math:`W(t)`. Moreover, the initial condition is :math:`\psi(t_0)`
    and :math:`W(t)` is expected to be absent before the initial time
    :math:`t_0`:

        :math:`W(t) = 0 \,\, \text{for} \,\, t \leq t_0`.

    If an energy :math:`E` is provided, this routine expects that
    the initial condition represents the scattering state :math:`\psi_{st}`
    that solves the time-independent Schrödinger equation

        :math:`H_0 \psi_{st} = E \psi_{st}, \, \psi_{st} = \psi(t_0)`.

    For numerical reasons, the evolution is then performed in the variable

        :math:`i \partial_t \bar{\psi} = (H_0 + W(t) - E) \bar{\psi} + W(t) \psi_{st}`,

    where

        :math:`\psi = (\bar{\psi} - \psi_{st}) e^{-i E (t - t_0)}`.

    See `J. Weston and X. Waintal, Phys. Rev. B 93, 134506 (2016)
    <https://arxiv.org/abs/1510.05967>`_.
    """
    def __init__(self, H0, W, psi_init, energy=None, params=None,
                 time_start=_common.time_start, time_name=_common.time_name,
                 kernel_type=kernels.default, solver_type=solvers.default,
                 work=None, inplace=False, tmax=None):
        r"""
        Parameters
        ----------
        H0 : array-like
            The static part of the Hamiltonian matrix, :math:`H_0`.
        W : callable or `None`
            Time-dependent part of the Hamiltonian matrix, :math:`W(t)`. Typically
            the object returned by `tkwant.system.extract_perturbation`.
        psi_init : array of complex
            The state :math:`\psi(t_0)` from which to start, defined over the
            central region.
        energy : float, optional
            If provided, then ``psi_init`` is assumed to be an eigenstate
            of energy :math:`E`. If the Hamiltonian represents an open
            quantum system with leads, then ``psi_init`` is
            assumed to be the projection of a scattering state at energy :math:`E`
            on to the central part of the system.
        params : dict, optional
            Extra arguments to pass to the time-dependent Hamiltonian
            function :math:`W(t)`, excluding time.
        time_start : float, optional
            The initial time :math:`t_0`. Default value is zero.
        time_name : str, optional
            The name of the time argument :math:`t`. Default name: *time*.
        kernel_type : `tkwant.onebody.solvers.default`, optional
            The kernel to calculate the right-hand-site of the
            Schrödinger equation.
        solver_type : `tkwant.onebody.solvers.default`, optional
            The solver used to evolve the wavefunction forward in time.
        work : `~numpy.ndarray` of complex, optional
            Workarray of size ``H0.shape[0]`` for performance and memory optimization.
        inplace : bool, optional
            If true and ``energy`` is not None,
            the term ``H0 - E`` is calculated inplace in order to save memory.
            If ``energy`` is None, ``inplace`` has no effect.
        tmax : float, optional
            Optional maximal time until when the boundary is valid.
        """

        # The size of the central scattering region. The central scattering
        # region can be smaller as the total size (kernel.size) of the system
        # in case of boundary conditions.
        syst_size = psi_init.size
        hamiltonian_size = H0.shape[0]

        if syst_size > hamiltonian_size:
            raise ValueError('initial condition size={} is larger than the '
                             'Hamiltonian matrix H0 size={}'
                             .format(syst_size, hamiltonian_size))

        # The perturbation W(t) must, if present, always match the size of the
        # central scattering region. The true leads are never time dependent.
        if W is not None:
            if not W.size == syst_size:
                raise ValueError('initial condition size={} must be equal '
                                 'to the perturbation W size={}'
                                 .format(syst_size, W.size))

        if work is not None:
            if not work.size == hamiltonian_size:
                raise ValueError('work array size={} must be equal '
                                 'Hamiltonian matrix H0 size={}'
                                 .format(work.size, hamiltonian_size))

        # transform initial psi to psibar and psi_st
        # note that the dgl is always solved in the variable psibar
        psibar = np.zeros((hamiltonian_size,), complex)

        if energy is not None:
            # starting from an arbitrary state, so we need
            # to be solving: H0 @ psi + W(t) @ psi
            psi_st = np.array(psi_init, complex)
        else:
            # we are starting from an eigenstate, so we need to
            # be solving: (H0 - E) @ psibar + W(t) @ (psibar + psi_st)
            # and psi = (psibar + psi_st) * exp(-1j * energy * time)
            psi_st = None
            psibar[:syst_size] = psi_init

        # get the object that will actually do the time stepping
        try:
            self.solver = solver_type(hamiltonian_size)
        except Exception:
            self.solver = solver_type
        if not self.solver.size == hamiltonian_size:
            raise ValueError('solver size={} must be equal to '
                             'Hamiltonian matrix H0 size={}'
                             .format(self.solver.size, hamiltonian_size))

        # kernel to evaluate the r.h.s. of the Schödinger eq.
        if not inplace and energy is not None:
            self.kernel = kernel_type(H0 - energy * sp.eye(hamiltonian_size),
                                      W, psi_st, work)
        else:
            self.kernel = kernel_type(H0, W, psi_st, work)
        self._inplace = True if (inplace and energy is not None) else False

        self.psibar = psibar
        self.psi_st = psi_st

        self._syst_size = syst_size

        self.time_name = time_name
        self.time_start = time_start

        self.time = time_start
        self.params = params
        self.energy = energy
        self.tmax = tmax
        self._perturb = None

    @classmethod
    def from_kwant(cls, syst, psi_init, boundaries=None, energy=None, params=None,
                   time_start=_common.time_start, time_name=_common.time_name,
                   kernel_type=kernels.default, solver_type=solvers.default,
                   perturbation_type=kernels.PerturbationInterpolator):
        """Set up a time-dependent onebody wavefunction from a kwant system.

        Parameters
        ----------
        syst : `kwant.builder.FiniteSystem`
            The low level system for which the wave functions are to be
            calculated.
        psi_init : array of complex
            The state from which to start, defined over the central region.
        boundaries : sequence of `~tkwant.leads.BoundaryBase`, optional
            The boundary conditions for each lead attached to ``syst``.
            Must be provided for a system with leads.
        energy : float, optional
            If provided, then ``psi_init`` is assumed to be an eigenstate
            of energy *E*. If ``syst`` has leads, then ``psi_init`` is
            assumed to be the projection of a scattering state at energy *E*
            on to the central part of the system.
        params : dict, optional
            Extra arguments to pass to the Hamiltonian of ``syst``, excluding time.
        time_start : float, optional
            The initial time :math:`t_0`. Default value is zero.
        time_name : str, optional
            The name of the time argument :math:`t`. Default name: *time*.
        kernel_type : `tkwant.onebody.solvers.default`, optional
            The kernel to calculate the right-hand-site of the
            Schrödinger equation.
        solver_type : `tkwant.onebody.solvers.default`, optional
            The solver used to evolve the wavefunction forward in time.
        perturbation_type : `tkwant.onebody.kernels.ExtractPerturbation`, optional
            Class to extract the time dependent perturbation :math:`W(t)`
            out of ``syst``.

        Returns
        -------
        wave_function : `tkwant.onebody.WaveFunction`
            A time-dependent onebody wavefunction at the initial time.
        """
        # add initial time to the params dict
        tparams = system.add_time_to_params(params, time_name=time_name,
                                            time=time_start, check_numeric_type=True)

        if isinstance(syst, kwant.system.System):
            extended_syst = make_extended_system(syst, solver_type, boundaries, tparams)
        elif isinstance(syst, system.ExtendedSystem):
            extended_syst = syst
            syst = extended_syst.syst
        else:
            raise TypeError('"syst" must be a finalized kwant system '
                            'or a "tkwant.system.ExtendedSystem"')

        syst_size = extended_syst.syst.site_ranges[-1][2]
        if not psi_init.size == syst_size:
            raise ValueError('Size of the initial condition={} does not match '
                             'the total number of orbitals in the central '
                             'system ={}'.format(psi_init.size, syst_size))

        H0 = extended_syst.H0
        work = extended_syst.work
        solver = extended_syst.solver
        tmax = extended_syst.tmax

        try:
            W = syst.time_dependent_perturbation(time_name, time_start, params)
        except AttributeError:
            W = perturbation_type(syst, time_name, time_start, params=params)

        return cls(H0, W, psi_init, energy, params,
                   time_start, time_name, kernel_type, solver, work, tmax=tmax)

    def psi(self):
        r"""Return the wavefunction :math:`\psi(t)` at the current time *t*.

        psi : `~numpy.ndarray`
            The wavefunction at ``time``. If this wavefunction is
            for a system with leads, then the wavefunction projected
            onto the central region is returned.
        """
        if self.energy is None:  # psi = psibar
            return self.psibar[:self._syst_size].copy()
        # else : psi = (psibar + psi_st) * exp(-i E (t - t0))
        return ((self.psibar[:self._syst_size] + self.psi_st) *
                exp(-1j * self.energy * (self.time - self.time_start)))

    def evolve(self, time):
        r"""
        Evolve the wavefunction :math:`\psi(t)` foreward in time up to
        :math:`t =` ``time``.

        Parameters
        ----------
        time : int or float
            time argument up to which the solver should be evolved
        """
        # `time` corresponds to the future time, to which we will evolve
        if time < self.time:
            raise ValueError('Cannot evolve backwards in time')
        if time == self.time:
            return
        if self.tmax is not None:
            if time > self.tmax:
                logger.warning('time > tmax; time= {}, tmax= {}'. format(time, self.tmax))

        # evolve forward in time
        if self._inplace:
            next_psibar = self.solver(self.psibar, self.time, time,
                                      self.kernel.rhs(self.energy))
        else:
            next_psibar = self.solver(self.psibar, self.time, time,
                                      self.kernel.rhs())

        # update internal state and return
        self.psibar, self.time = next_psibar, time

    def evaluate(self, observable):
        r"""
        Evaluate the expectation value of an operator at the current time *t*.

        For an operator :math:`\hat{O}` the expectation value is
        :math:`O(t) = \langle \psi(t) | \hat{O} |\psi(t) \rangle`.

        Parameters
        ----------
        observable : callable or `kwant.operator`
            An operator :math:`\hat{O}` to evaluate the expectation value.
            Must have the calling signature of `kwant.operator`.

        Returns
        -------
        result : numpy array
            The expectation value :math:`O(t)` of ``observable``.
        """
        if _operator_bound(observable):
            raise ValueError("Operator must not use pre-bind values")

        tparams = system.add_time_to_params(self.params, time_name=self.time_name,
                                            time=self.time)
        return observable(self.psi(), params=tparams)

    def add_perturbation(self, qt):
        """Add a time-dependent perturbation to the Hamiltonian

         H(t) = H_0 + W(t)

         it is modified to

         H(t) = H_0 + W(t) + Q(t)
        """
        if self._perturb is None:
            self._perturb = _PerturbationWrap(self.kernel.W)
        self._perturb.set(qt)
        self.kernel.set_W(self._perturb)


class ScatteringStates(collections.abc.Iterable):
    """Calculate time-dependent wavefunctions starting in an equilibrium scattering state"""

    def __init__(self, syst, energy, lead, tmax=None, params=None, spectra=None,
                 boundaries=None, equilibrium_solver=kwant.wave_function,
                 wavefunction_type=WaveFunction.from_kwant):
        """
        Parameters
        ----------
        syst : `kwant.builder.FiniteSystem`
            The low level system for which the wave functions are to be
            calculated.
        energy : float
            Energy of the scattering eigenstate.
        lead : int
            Lead index to construct the scattering state.
        tmax : float, optional
            The maximum time up to which to simulate. Sets the boundary conditions
            such that they are guaranteed to be correct up to 'tmax'. Mutually
            exclusive with 'boundaries'.
        params : dict, optional
            Extra arguments to pass to the Hamiltonian of ``syst``, excluding time.
        spectra : sequence of `kwantspectrum.spectrum`, optional
            Energy dispersion :math:`E_n(k)` for the leads. Must have
            the same length as ``syst.leads``. Required only if
            no ``boundaries`` are provided. If needed but not present,
            it will be calculated on the fly from `syst.leads`.
        boundaries : sequence of `~tkwant.leads.BoundaryBase`, optional
            The boundary conditions for each lead attached to ``syst``. Mutually
            exclusive with 'tmax'.
        equilibrium_solver : `kwant.wave_function`, optional
            Solver for initial equilibrium scattering problem.
        wavefunction_type : `WaveFunction`, optional
            One-body time-dependent wave function. Name of the time argument and
            initial time are taken from this class. If this is not possible,
            default values are used as a fallback.

        Notes
        -----
        An instance of this class behaves like a sequence of `WaveFunction`
        instances. The index in the sequence corresponds to the scattering mode.
        The name of the time argument (`time_name`) and the initial time
        of the evolution (`time_start`) are taken from the default
        values of the `WaveFunction.__init__` method. Changing the default
        values by partial prebind (e.g. via functools) is possible.
        """

        if isinstance(syst, kwant.system.System):
            extended_syst = None
        elif isinstance(syst, system.ExtendedSystem):
            if boundaries is not None:
                logger.warning('boundaries will not be used for extended system')
            extended_syst = syst
            syst = extended_syst.syst
        else:
            raise TypeError('"syst" must be a finalized kwant system '
                            'or a "tkwant.system.ExtendedSystem"')
        if not syst.leads:
            raise AttributeError("system has no leads.")
        if tmax is not None and boundaries is not None:
            raise ValueError("'boundaries' and 'tmax' are mutually exclusive.")
        if not _common.is_type(energy, 'real_number'):
            raise TypeError('energy must be a real number')
        if not _common.is_type(lead, 'integer'):
            raise TypeError('lead index must be an integer')
        if lead >= len(syst.leads):
            raise ValueError("lead index must be smaller than {}.".format(len(syst.leads)))

        # get initial time and time argument name from the onebody wavefunction
        time_name, time_start, _ = _get_time_name_from_wavefunction(wavefunction_type)

        # add initial time to the params dict
        tparams = system.add_time_to_params(params, time_name=time_name,
                                            time=time_start, check_numeric_type=True)

        if extended_syst is None:
            solver_type = _common.get_default_function_argument(wavefunction_type,
                                                                'solver_type')
            if boundaries is None:
                if spectra is None:
                    spectra = kwantspectrum.spectra(syst.leads, params=tparams)
                boundaries = leads.automatic_boundary(spectra, tmax)
            extended_syst = make_extended_system(syst, solver_type, boundaries, tparams)

        scattering_states = equilibrium_solver(syst, energy=energy, params=tparams)
        self._psi_st = scattering_states(lead)
        self._wavefunction = functools.partial(wavefunction_type,
                                               syst=extended_syst, params=params)
        self.energy = energy
        self.lead = lead

    def __getitem__(self, mode):
        """Return a time-dependent wavefunction for the corresponding scattering mode.

        Parameters
        ----------
        mode : int
            Mode index of the scattering wavefunction.

        Returns
        -------
        wave_function : `WaveFunction`
            The time-dependent wave function starting in the scattering mode.
            Two additional attributes, ``lead`` and ``mode`` are added
            to the wave function.
        """

        if not _common.is_type(mode, 'integer'):
            raise TypeError('mode index must be an integer')
        if mode >= len(self._psi_st):
            raise KeyError('no open scattering mode={}; '
                           'number of open modes={} for energy={} and lead={}'
                           .format(self.lead, mode, self.energy, len(self._psi_st)))
        psi = self._wavefunction(psi_init=self._psi_st[mode], energy=self.energy)
        psi.lead = self.lead
        psi.mode = mode
        return psi

    def __len__(self):
        """Return the number of open scattering modes."""
        return len(self._psi_st)

    def __iter__(self):
        """Return an iterable sequence of all open modes."""
        return (self[mode] for mode in range(len(self)))



# -*- coding: utf-8 -*-
# Copyright 2016-2019 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.
"""Tools for testing parallel code."""

import os
import inspect
import subprocess

__all__ = ["wrap"]


def _format_path(func):
    """Return a working path and module"""
    file = inspect.getfile(func)
    path = os.path.abspath(file)
    module = func.__module__

    for i in range(module.count(".")):
        path = os.path.split(path)[0]

    return path, module


def wrap(num_processes=1):
    """Return a decorator to run function with mpirun.
    Function must not take argument.

    num_processes: int, numer of MPI core
    parents: int, number of parent to include in working directory
    """

    def wrapper(func):
        path, module = _format_path(func)
        command = ["mpirun", "-n", str(num_processes), "-wdir", path, "python3", "-c",
                   "import {0} as module; module.{1}.original()"
                   .format(module, func.__name__)]

        def new_func():
            subprocess.run(command, check=True)

        new_func.original = func

        return new_func

    return wrapper



import tkwant
import kwant
import numpy as np
import matplotlib.pyplot as plt
import functools as ft


def am_master():
    """Return true for the MPI master rank"""
    return tkwant.mpi.get_communicator().rank == 0


def side_color_func(site):
    return 'r' if electrode_shape(site.pos) else ('k' if site.family == a else 'w')


def onsite_potential(site, time):
    # Time-dependent potential (static part + V(t))
    return 0.001 * np.exp(- 0.01 * (time - 20)**2)


def circle(pos, x0, y0, r):
    x, y = pos
    return (x - x0)**2 + (y - y0)**2 < r**2


def electrode_shape(pos):
    x, y = pos
    return (-0.1 < x < 0.1) and (1 < y < 3)


def lead_shape(site):
    x, y = site.pos
    return (-1.5 < y < 1.0)


def make_system():
    # Define the graphene lattice.
    lat = kwant.lattice.honeycomb(a=1, norbs=1)
    a, b = lat.sublattices

    # Create graphene model.
    model = kwant.Builder(kwant.TranslationalSymmetry(
        lat.vec((1, 0)), lat.vec((0, 1))))
    model[[a(0, 0), b(0, 0)]] = 0
    model[lat.neighbors()] = -1

    # Central scattering region.
    funs = [ft.partial(circle, x0=0, y0=0, r=3.3)]
    syst = kwant.Builder()
    syst.fill(model, lambda site: any(f(site.pos) for f in funs), a(0, 0))
    syst.eradicate_dangling()
    syst[lat.shape(electrode_shape, (0, 2))] = onsite_potential

    # Define the lead using a trick to avoid ugly diagonal lead interfaces.
    sym = kwant.TranslationalSymmetry(lat.vec((1, 0)))
    sym.add_site_family(a, other_vectors=[(-1, 2)])
    sym.add_site_family(b, other_vectors=[(-1, 2)])

    lead = kwant.Builder(sym)
    lead.fill(model, lead_shape, a(0, 0))
    syst.attach_lead(lead)

    return syst


chemical_potential = -2.5
times = [10, 15, 20, 25, 30]

syst = make_system()

# plot the system
lat = kwant.lattice.honeycomb(a=1, norbs=1)
a, b = lat.sublattices
if am_master():
    kwant.plot(syst, site_lw=0.1, site_color=side_color_func,
               lead_site_lw=0, lead_color='grey')

syst = syst.finalized()

density_operator = kwant.operator.Density(syst)
occupations = tkwant.manybody.lead_occupation(chemical_potential)

# Initialize the time-dependent manybody state. Here we use a lower
# accuracy for the initial adaptive refinement (at time=0) to speed up the calculation.
# For a realistic simulation, set rtol and atol to smaller values.
state = tkwant.manybody.State(syst, max(times), occupations, refine=False)
state.refine_intervals(rtol=1.5E-2, atol=1.5E-2)


# simulation
density0 = state.evaluate(density_operator)
densities = []
for time in times:
    state.evolve(time)
    # For a realistic simulation, one must set rtol and atol to smaller values,
    # or remove both arguments completely, to converge the manybody integral.
    state.refine_intervals(rtol=0.1, atol=0.1)
    density = state.evaluate(density_operator)
    if am_master():
        densities.append(density - density0)

# plot the result
if am_master():
    normalization = 1 / np.max(densities)
    for time, density in zip(times, densities):
        density *= normalization
        print('time={}'.format(time))
        kwant.plotter.density(syst, density, relwidth=0.15,
                              cmap='PuOr', vmin=-0.5, vmax=0.5)



import pickle
import pylab

import numpy as np

import matplotlib.pyplot as plt


# --- plot settings
markersize = 14

fac = 8
pylab.rcParams['figure.figsize'] = (fac * np.sqrt(2), fac)

plt.rcParams.update({'font.size': 16})
plt.rc('text', usetex=True)
plt.rc('font', family='serif')

colors = {0: '#e15759', 1: '#f28e2b', 2: '#4e79a7', 3: '#76b7b2', 4: '#59a14f',
          5: '#edc948', 6: '#b07aa1', 7: '#ff9da7', 8: '#9c755f', 9: '#bab0ac'}


# --- read the data
data = pickle.load(open('./siam_sudden_switch_data.npy', 'rb'))

times = data['times']
t0 = data['t0']

gless_tt0_tkwant = data['gless_tt0_tkwant']
gless_tt0_flatband = data['gless_tt0_flatband']
gless_tt0_flatband_num = data['gless_tt0_flatband_num']
gless_tt0_riwar = data['gless_tt0_riwar']
n_eq_tkwant = data['dens_tkwant_equilibrium']

gless_tt_tkwant = data['gless_tt_tkwant']
n_t_tkwant = data['n_t_tkwant']
gless_tt_flatband = data['gless_tt_flatband']
gless_tt_flatband_num = data['gless_tt_flatband_num']
gless_tt_riwar = data['gless_tt_riwar']
n_eq_flatband = data['dens_flatband_equilibrium']


# --- plot figure
fig, ax = plt.subplots(1, 2)
fig.set_size_inches(16, 5)

plt.suptitle(r'check numerical fourier transform vs analytical form $\gamma={}, \epsilon={}, t_0={}$'.
             format(data['gamma'], data['epsilon'], t0), size=22)

ax[0].plot(times, gless_tt0_flatband_num.real, color=colors[0], label=r'$\Re G^<(t, t_0)$ flatband num Fourier')
ax[0].plot(times, gless_tt0_flatband_num.imag, color=colors[1], label=r'$\Im G^<(t, t_0)$ flatband num Fourier')
ax[0].plot(times, gless_tt0_flatband.real, '--', color=colors[2], label=r'$\Re G^<(t, t_0)$ flatband')
ax[0].plot(times, gless_tt0_flatband.imag, '--', color=colors[3], label=r'$\Im G^<(t, t_0)$ flatband')

ax[0].set_xlabel(r'time $t$')
ax[0].set_ylabel(r'$G^<(t, t_0)$')
ax[0].set_xlim(0, times[-1])
ax[0].legend(fontsize=14)

ax[1].plot(times, gless_tt_flatband_num.imag, color=colors[0], label=r'$- i G^<(t, t)$ flatband num Fourier')
ax[1].plot(times, gless_tt_flatband.imag, '--', color=colors[2], label=r'$- i G^<(t, t)$ flatband')
ax[1].plot([0, times[-1]], [n_eq_flatband] * 2, linestyle='dotted', color=colors[1], label=r'$n_{eq}$ flatband')

ax[1].set_xlabel(r'time $t$')
ax[1].set_ylabel(r'$n(t)$')
ax[1].set_xlim(0, times[-1])
ax[1].legend(fontsize=14)

plt.show()

fig, ax = plt.subplots(1, 2)
fig.set_size_inches(16, 5)

plt.suptitle(r'compare result against formula from riwar/schmidt $\gamma={}, \epsilon={}, t_0={}$'.
             format(data['gamma'], data['epsilon'], t0), size=22)

ax[0].plot(times, gless_tt0_riwar.real, color=colors[0], label=r'$\Re G^<(t, t_0)$ flatband riwar')
ax[0].plot(times, gless_tt0_riwar.imag, color=colors[1], label=r'$\Im G^<(t, t_0)$ flatband riwar')
ax[0].plot(times, gless_tt0_flatband.real, '--', color=colors[2], label=r'$\Re G^<(t, t_0)$ flatband')
ax[0].plot(times, gless_tt0_flatband.imag, '--', color=colors[3], label=r'$\Im G^<(t, t_0)$ flatband')

ax[0].set_xlabel(r'time $t$')
ax[0].set_ylabel(r'$G^<(t, t_0)$')
ax[0].set_xlim(0, times[-1])
ax[0].legend(fontsize=14)

ax[1].plot(times, gless_tt_riwar.imag, color=colors[0], label=r'$- i G^<(t, t)$ flatband riwar')
ax[1].plot(times, gless_tt_riwar.imag, '--', color=colors[2], label=r'$- i G^<(t, t)$ flatband')
ax[1].plot([0, times[-1]], [n_eq_flatband] * 2, linestyle='dotted', color=colors[1], label=r'$n_{eq}$ flatband')

ax[1].set_xlabel(r'time $t$')
ax[1].set_ylabel(r'$n(t)$')
ax[1].set_xlim(0, times[-1])
ax[1].legend(fontsize=14)

plt.show()


fig, ax = plt.subplots(1, 2)
fig.set_size_inches(16, 5)

ax[0].plot(times, gless_tt0_tkwant.real, color=colors[0], label=r'$\Re G^<(t, t_0)$ Tkwant')
ax[0].plot(times, gless_tt0_tkwant.imag, color=colors[1], label=r'$\Im G^<(t, t_0)$ Tkwant')
ax[0].plot(times, gless_tt0_flatband.real, linestyle='dotted', lw=3, color=colors[2], label=r'$\Re G^<(t, t_0)$ flatband')
ax[0].plot(times, gless_tt0_flatband.imag, linestyle='dotted', lw=3, color=colors[3], label=r'$\Im G^<(t, t_0)$ flatband')
ax[0].plot(t0, n_eq_tkwant, 'x', markersize=12, color=colors[1], label=r'$n_{eq}$ Tkwant')
ax[0].plot(t0, n_eq_flatband, 'x', markersize=12, color=colors[3], label=r'$n_{eq}$ flatband')

ax[0].set_xlabel(r'time $t$')
ax[0].set_ylabel(r'$G^<(t, t_0)$')
ax[0].set_xlim(0, times[-1])
ax[0].legend(fontsize=14)

ax[1].plot(times, gless_tt_tkwant.imag, color=colors[0], label=r'$- i G^<(t, t)$ Tkwant')
ax[1].plot(times, gless_tt_flatband.imag, color=colors[1], label=r'$- i G^<(t, t)$ flatband')
ax[1].plot(times, n_t_tkwant, linestyle='dotted', lw=3, color=colors[2], label=r'$n(t)$ Tkwant')
ax[1].plot([0, times[-1]], [n_eq_tkwant] * 2, linestyle='dashed', color=colors[2], label=r'$n_{eq}$ Tkwant')
ax[1].plot([0, times[-1]], [n_eq_flatband] * 2, linestyle='dashed', color=colors[1], label=r'$n_{eq}$ flatband')

ax[1].set_xlabel(r'time $t$')
ax[1].set_ylabel(r'$n(t)$')
ax[1].set_xlim(0, times[-1])

ax[1].legend(fontsize=14)

plt.savefig("fig_siam_sudden_switch.png", bbox_inches='tight')
plt.show()



# -*- coding: utf-8 -*-
# Copyright 2016-2022 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.
"""Tools for extracting time-dependent parts of Kwant systems."""

import numpy as np
import inspect
import scipy.sparse as sp
import kwant

from .import _common

__all__ = ['add_time_to_params', 'ExtendedSystem',
           'is_time_dependent_function', 'orb_range', 'siteId', 'Hamiltonian']


def add_time_to_params(params, time_name, time, check_numeric_type=False):
    """Add a ``time_name: time`` key-value pair to a ``params`` dict.

    Parameters
    ----------
    params : dict or None
        Input dict. Can be empty or None. ``params`` must
        not contain the ``time_name`` key - this will raise a `KeyError`.
    time_name : obj
        A dict key to refer to the "time".
    time : obj
        The actual value of "time".
    check_numeric_type : bool, optional
        If true, check if the `time` argument is a finite real number.
        By default, no check is performed.

    Returns
    -------
    tparams : dict
        A copy of the input dict ``params`` that contains an additional
        ``time_name: time`` key-value pair.
    """
    if check_numeric_type:
        if not _common.is_type(time, 'real_number', require_finite=True):
            raise TypeError('time must be a finite real number')

    if params is None:
        tparams = {}
    else:
        if time_name in params:
            raise KeyError("'params' must not contain {}".format(time_name))
        tparams = params.copy()
    tparams[time_name] = time
    return tparams


class ExtendedSystem:
    """Hamiltonian matrix of central system + boundary conditions.

    Parameters
    ----------
    hamiltonian : `~scipy.sparse.coo_matrix`
        The Hamiltonian matrix of the central system + boundary conditions.
    boundary_slices : sequence of `slice`
        Slices into ``hamiltonian`` that project onto the boundary conditions.
    syst : `~kwant.system.FiniteSystem`
        The original kwant system.
    tmax : float, optional
        Optional maximal time until when the boundary is valid.
    """
    def __init__(self, syst, H0, boundary_slices, solver, work, tmax=None):

        self.syst = syst
        self.H0 = H0
        self.boundary_slices = boundary_slices
        self.solver = solver
        self.work = work
        self.tmax = tmax


def _hamiltonian_with_boundaries(syst, boundaries, params):
    r"""Generate the Hamiltonian with boundary conditions attached.

    Only generate the time-independent part of the Hamiltonian.
    The boundary conditions are represented by matrices :math:`h_n`
    (one per lead) that will form a direct sum with the Hamiltonian
    of the central system, :math:`H_S`:

    .. math:: H_{tot} = H_S \oplus h_0 \oplus h_1 \oplus \cdots

    In addition, coupling terms given by the lead inter-cell
    hoppings will be added between the added boundary conditions
    and the corresponding lead interface in the central system
    Hamiltonian.

    Parameters
    ----------
    syst : `~kwant.system.FiniteSystem`
        System with leads attached.
    boundaries : sequence of `~tkwant.leads.BoundaryBase`
        The boundary conditions to attach; one per lead.
    params : dict, optional
        Extra arguments to pass to the Hamiltonian of ``syst`` at the
        initial time. Must include the time argument explicitly if
        Hamiltonian is time dependent.

    Returns
    -------
    Hamiltonian H0 and first boundary slice to the lead.
    """

    if not len(syst.leads) == len(boundaries):
        raise ValueError('Number of leads= {} does not match '
                         'the number of boundaries provided= {}'
                         .format(len(syst.leads), len(boundaries)))

    # generate time-independent Hamiltonian and boundary condition
    # matrices and glue them together
    boundaries = [bc(lead, params) for lead, bc in zip(syst.leads, boundaries)]
    h_0 = syst.hamiltonian_submatrix(params=params, sparse=True)
    h_tot = sp.block_diag([h_0] + [bdy.hamiltonian for bdy in boundaries],
                          format='lil')

    boundary_slices = []

    # couple central system to boundary conditions and vice versa
    orb_offset = h_0.shape[0]
    for lead, lead_iface, boundary in zip(syst.leads, syst.lead_interfaces,
                                          boundaries):
        V = lead.inter_cell_hopping(params=params)
        # slices over lead interface orbitals within the lead
        num_iface_sites = lead.graph.num_nodes - lead.cell_size
        V_slices = [slice(*orb_range(lead, s)) for s in range(num_iface_sites)]
        # slices going *to* and *from* the boundary conditions/central system
        to_slices = [slice(*(s + orb_offset)) for s in boundary.to_slices]
        from_slices = [slice(*(s + orb_offset)) for s in boundary.from_slices]

        # iterate through all sites in the interface -- they have to
        # be treated separately, as interface sites in `syst` are not
        # necessarily grouped consecutively.
        for syst_site, V_slice in zip(lead_iface, V_slices):
            syst_slice = slice(*orb_range(syst, syst_site))
            # coupled *to* system, *from* boundary conditions
            for to_slice in to_slices:
                h_tot[syst_slice, to_slice] =\
                    V[:, V_slice].conjugate().transpose()
            # couple *from* system, *to* boundary conditions
            for from_slice in from_slices:
                h_tot[from_slice, syst_slice] = V[:, V_slice]
                pass

        # remember which orbitals correspond to this boundary condition
        boundary_slices.append(slice(orb_offset,
                                     orb_offset + boundary.hamiltonian.shape[0]))
        # now point to the start of the next boundary condition block
        orb_offset += boundary.hamiltonian.shape[0]

    return h_tot.tocsr(), boundary_slices


def is_time_dependent_function(obj, time_name):
    """Return `True` if `obj` is callable and has a time argument"""
    return callable(obj) and time_name in str(inspect.signature(obj))


def orb_range(syst, site):
    """Return the first orbital of this and the next site.

    Parameters
    ----------
    syst : `~kwant.system.System`
    site : int

    Returns
    -------
    pair of integers
    """
    assert 0 <= site < syst.graph.num_nodes
    if syst.site_ranges is None:
        raise RuntimeError('Number of orbitals not defined.\n'
                           'Declare the number of orbitals using the `norbs` '
                           'keyword argument when constructing site families')
    # Calculate the index of the run that contains the site.
    column = np.asarray(syst.site_ranges)[:, 0]
    run_idx = np.searchsorted(column, site, 'right') - 1
    # calculate the slice
    first_site, norbs, orb_offset = syst.site_ranges[run_idx]
    orb = orb_offset + (site - first_site) * norbs
    return orb, orb + norbs


class siteId:
    """Get the site index from the lattice position

    For instance, building a kwant system with:
        lat = kwant.lattice.square(a=1)
        syst = kwant.Builder()
        syst[lat(3, 4)] = 42
        syst = syst.finalized()
    the second line of:
        idx = siteId(syst, lat)
        idx(3, 4)
    will return the index of the site (whose onsite
    element has been set to 42 in the lines above) in kwant ordering.
    """

    def __init__(self, syst, lat):

        if not isinstance(syst, kwant.system.System):
            raise TypeError("'syst' must be an instance of 'kwant.system.System'")

        self._syst = syst
        self._lat = lat

    def __call__(self, *lat_coord, orbital=0):
        """Return the site index for a given lattice index.

        Parameters
        ----------
        lat_coord : tuple of int
            Lattice coordinates to refer to a lattice positions.
        orbital: int
            Orbital index

        Returns
        -------
        site_id : int
            The site index in kwant ordering.
        """
        site = self._lat(*lat_coord)
        site_id = self._syst.id_by_site[site]
        group = None
        for elem, (xx, _, _) in enumerate(self._syst.site_ranges):
            if xx > site_id:
                group = elem - 1
                break
        try:
            first_site, norbs, orb_offset = self._syst.site_ranges[group]
        except TypeError:
            raise ValueError('Group position not found')
        if not 0 <= orbital < norbs:
            raise ValueError("'0 <= orbital < norbs'; orbital={}, norbs={}"
                             .format(orbital, norbs))
        return (site_id - orb_offset) * norbs + orb_offset + orbital


def _site_to_matpos(syst, site):
    pos = syst.id_by_site[site]
    group = None
    for elem, (xx, _, _) in enumerate(syst.site_ranges):
        if xx > pos:
            group = elem - 1
            break
    try:
        first_site, norbs, orb_offset = syst.site_ranges[group]
    except TypeError:
        raise ValueError('Group position not found')
    return (pos - orb_offset) * norbs + orb_offset, norbs


class Hamiltonian:
    """Construct a Hamiltonian matrix for a single site or hopping.

    Constructing a kwant system with builder is simple.
    However, once a system is finalized, Kwant has put onsite elements and
    hoppings in some ordering and it is not evident to get the positions correctly.
    This routine constructs a Hamiltonian matrix given directly the site,
    or the hopping position. On initialization, the Hamiltonian matrix is
    construced. Calling the get method, one needs to provide the matrix
    subblock, which is inserted at the site or hopping position in
    the full Hamiltonian matrix.
    """
    def __init__(self, syst, site=None, hopping=None):
        if site is None and hopping is None:
            raise ValueError('either a site or a hopping must be present')
        if site is not None and hopping is not None:
            raise ValueError('site and hopping are mutually exclusive')
        if site is not None:
            pos1, norbs1 = _site_to_matpos(syst, site)
            pos2, norbs2 = pos1, norbs1
            self._is_hopping = False
        else:
            assert len(hopping) == 2
            pos1, norbs1 = _site_to_matpos(syst, hopping[0])
            pos2, norbs2 = _site_to_matpos(syst, hopping[1])
            self._is_hopping = True
        self._size = syst.site_ranges[-1][2]  # total hamiltonian size
        row = []
        col = []
        for i in range(norbs1):
            for j in range(norbs2):
                row.append(pos1 + i)
                col.append(pos2 + j)
        if self._is_hopping:
            self._row = np.asarray(row + col)
            self._col = np.asarray(col + row)
        else:
            self._row = np.asarray(row)
            self._col = np.asarray(col)
        return

    def get(self, submat):
        data = np.asarray(submat).flatten()
        if self._is_hopping:
            data = np.append(data, np.asarray(submat).T.conjugate().flatten())
        qt = sp.coo_matrix((data, (self._row, self._col)), dtype=complex,
                           shape=(self._size, self._size))
        qt.eliminate_zeros()
        return qt
